######################################################################
#
# File: test/unit/v1/deps.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

# TODO: This module is used in old-style unit tests, written separately for v0 and v1.
#  It will be removed when all test are rewritten for the new style, like e.g. test/unit/sync/.

# configured by pytest using `--api` option
# check test/unit/conftest.py:pytest_configure for details
from apiver_deps import *
